# Copyright (C) 2014 Daniel Lee <lee.daniel.1986@gmail.com>
#
# This file is part of StereoVision.
#
# StereoVision is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# StereoVision is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with StereoVision.  If not, see <http://www.gnu.org/licenses/>.

"""
Classes for calibrating homemade stereo cameras.

Classes:

    * ``StereoCalibration`` - Calibration for stereo camera
    * ``StereoCalibrator`` - Class to calibrate stereo camera with

.. image:: classes_calibration.svg
"""

import os

import cv2

import numpy as np
from stereovision.exceptions import ChessboardNotFoundError


class StereoCalibration(object):

    """
    A stereo camera calibration.

    The ``StereoCalibration`` stores the calibration for a stereo pair. It can
    also rectify pictures taken from its stereo pair.
    """

    def __str__(self):
        output = ""
        for key, item in self.__dict__.items():
            output += key + ":\n"
            output += str(item) + "\n"
        return output

    def _copy_calibration(self, calibration):
        """Copy another ``StereoCalibration`` object's values."""
        for key, item in calibration.__dict__.items():
            self.__dict__[key] = item

    def _interact_with_folder(self, output_folder, action):
        """
        Export/import matrices as *.npy files to/from an output folder.

        ``action`` is a string. It determines whether the method reads or writes
        to disk. It must have one of the following values: ('r', 'w').
        """
        if not action in ('r', 'w'):
            raise ValueError("action must be either 'r' or 'w'.")
        for key, item in self.__dict__.items():
            if isinstance(item, dict):
                for side in ("left", "right"):
                    filename = os.path.join(output_folder,
                                            "{}_{}.npy".format(key, side))
                    if action == 'w':
                        np.save(filename, self.__dict__[key][side])
                    else:
                        self.__dict__[key][side] = np.load(filename)
            else:
                filename = os.path.join(output_folder, "{}.npy".format(key))
                if action == 'w':
                    np.save(filename, self.__dict__[key])
                else:
                    self.__dict__[key] = np.load(filename)

    def __init__(self, calibration=None, input_folder=None):
        """
        Initialize camera calibration.

        If another calibration object is provided, copy its values. If an input
        folder is provided, load ``*.npy`` files from that folder. An input
        folder overwrites a calibration object.
        """
        #: Camera matrices (M)
        self.cam_mats = {"left": None, "right": None}
        #: Distortion coefficients (D)
        self.dist_coefs = {"left": None, "right": None}
        #: Rotation matrix (R)
        self.rot_mat = None
        #: Translation vector (T)
        self.trans_vec = None
        #: Essential matrix (E)
        self.e_mat = None
        #: Fundamental matrix (F)
        self.f_mat = None
        #: Rectification transforms (3x3 rectification matrix R1 / R2)
        self.rect_trans = {"left": None, "right": None}
        #: Projection matrices (3x4 projection matrix P1 / P2)
        self.proj_mats = {"left": None, "right": None}
        #: Disparity to depth mapping matrix (4x4 matrix, Q)
        self.disp_to_depth_mat = None
        #: Bounding boxes of valid pixels
        self.valid_boxes = {"left": None, "right": None}
        #: Undistortion maps for remapping
        self.undistortion_map = {"left": None, "right": None}
        #: Rectification maps for remapping
        self.rectification_map = {"left": None, "right": None}
        if calibration:
            self._copy_calibration(calibration)
        elif input_folder:
            self.load(input_folder)

    def load(self, input_folder):
        """Load values from ``*.npy`` files in ``input_folder``."""
        self._interact_with_folder(input_folder, 'r')

    def export(self, output_folder):
        """Export matrices as ``*.npy`` files to an output folder."""
        if not os.path.exists(output_folder):
            os.makedirs(output_folder)
        self._interact_with_folder(output_folder, 'w')

    def rectify(self, frames):
        """
        Rectify frames passed as (left, right) pair of OpenCV Mats.

        Remapping is done with nearest neighbor for speed.
        """
        new_frames = []
        for i, side in enumerate(("left", "right")):
            new_frames.append(cv2.remap(frames[i],
                                        self.undistortion_map[side],
                                        self.rectification_map[side],
                                        cv2.INTER_NEAREST))
        return new_frames


class StereoCalibrator(object):

    """A class that calibrates stereo cameras by finding chessboard corners."""

    def _get_corners(self, image):
        """Find subpixel chessboard corners in image."""
        temp = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
        ret, corners = cv2.findChessboardCorners(temp,
                                                 (self.rows, self.columns))
        if not ret:
            raise ChessboardNotFoundError("No chessboard could be found.")
        cv2.cornerSubPix(temp, corners, (11, 11), (-1, -1),
                         (cv2.TERM_CRITERIA_MAX_ITER + cv2.TERM_CRITERIA_EPS,
                          30, 0.01))
        return corners

    def _show_corners(self, image, corners):
        """Show chessboard corners found in image."""
        temp = image
        cv2.drawChessboardCorners(temp, (self.rows, self.columns), corners,
                                  True)
        window_name = "Chessboard"
        cv2.imshow(window_name, temp)
        if cv2.waitKey(0):
            cv2.destroyWindow(window_name)

    def __init__(self, rows, columns, square_size, image_size):
        """
        Store variables relevant to the camera calibration.

        ``corner_coordinates`` are generated by creating an array of 3D
        coordinates that correspond to the actual positions of the chessboard
        corners observed on a 2D plane in 3D space.
        """
        #: Number of calibration images
        self.image_count = 0
        #: Number of inside corners in the chessboard's rows
        self.rows = rows
        #: Number of inside corners in the chessboard's columns
        self.columns = columns
        #: Size of chessboard squares in cm
        self.square_size = square_size
        #: Size of calibration images in pixels
        self.image_size = image_size
        pattern_size = (self.rows, self.columns)
        corner_coordinates = np.zeros((np.prod(pattern_size), 3), np.float32)
        corner_coordinates[:, :2] = np.indices(pattern_size).T.reshape(-1, 2)
        corner_coordinates *= self.square_size
        #: Real world corner coordinates found in each image
        self.corner_coordinates = corner_coordinates
        #: Array of real world corner coordinates to match the corners found
        self.object_points = []
        #: Array of found corner coordinates from calibration images for left
        #: and right camera, respectively
        self.image_points = {"left": [], "right": []}

    def add_corners(self, image_pair, show_results=False):
        """
        Record chessboard corners found in an image pair.

        The image pair should be an iterable composed of two CvMats ordered
        (left, right).
        """
        side = "left"
        self.object_points.append(self.corner_coordinates)
        for image in image_pair:
            corners = self._get_corners(image)
            if show_results:
                self._show_corners(image, corners)
            self.image_points[side].append(corners.reshape(-1, 2))
            side = "right"
            self.image_count += 1

    def calibrate_cameras(self):
        """Calibrate cameras based on found chessboard corners."""
        criteria = (cv2.TERM_CRITERIA_MAX_ITER + cv2.TERM_CRITERIA_EPS,
                    100, 1e-5)
        flags = (cv2.CALIB_FIX_ASPECT_RATIO + cv2.CALIB_ZERO_TANGENT_DIST +
                 cv2.CALIB_SAME_FOCAL_LENGTH)
        calib = StereoCalibration()
        (calib.cam_mats["left"], calib.dist_coefs["left"],
         calib.cam_mats["right"], calib.dist_coefs["right"],
         calib.rot_mat, calib.trans_vec, calib.e_mat,
         calib.f_mat) = cv2.stereoCalibrate(self.object_points,
                                            self.image_points["left"],
                                            self.image_points["right"],
                                            self.image_size,
                                            criteria=criteria,
                                            flags=flags)[1:]
        (calib.rect_trans["left"], calib.rect_trans["right"],
         calib.proj_mats["left"], calib.proj_mats["right"],
         calib.disp_to_depth_mat, calib.valid_boxes["left"],
         calib.valid_boxes["right"]) = cv2.stereoRectify(calib.cam_mats["left"],
                                                      calib.dist_coefs["left"],
                                                      calib.cam_mats["right"],
                                                      calib.dist_coefs["right"],
                                                      self.image_size,
                                                      calib.rot_mat,
                                                      calib.trans_vec,
                                                      flags=0)
        for side in ("left", "right"):
            (calib.undistortion_map[side],
             calib.rectification_map[side]) = cv2.initUndistortRectifyMap(
                                                        calib.cam_mats[side],
                                                        calib.dist_coefs[side],
                                                        calib.rect_trans[side],
                                                        calib.proj_mats[side],
                                                        self.image_size,
                                                        cv2.CV_32FC1)
        # This is replaced because my results were always bad. Estimates are
        # taken from the OpenCV samples.
        width, height = self.image_size
        focal_length = 0.8 * width
        calib.disp_to_depth_mat = np.float32([[1, 0, 0, -0.5 * width],
                                              [0, -1, 0, 0.5 * height],
                                              [0, 0, 0, -focal_length],
                                              [0, 0, 1, 0]])
        return calib

    def check_calibration(self, calibration):
        """
        Check calibration quality by computing average reprojection error.

        First, undistort detected points and compute epilines for each side.
        Then compute the error between the computed epipolar lines and the
        position of the points detected on the other side for each point and
        return the average error.
        """
        sides = "left", "right"
        which_image = {sides[0]: 1, sides[1]: 2}
        undistorted, lines = {}, {}
        for side in sides:
            undistorted[side] = cv2.undistortPoints(
                         np.concatenate(self.image_points[side]).reshape(-1,
                                                                         1, 2),
                         calibration.cam_mats[side],
                         calibration.dist_coefs[side],
                         P=calibration.cam_mats[side])
            lines[side] = cv2.computeCorrespondEpilines(undistorted[side],
                                              which_image[side],
                                              calibration.f_mat)
        total_error = 0
        this_side, other_side = sides
        for side in sides:
            for i in range(len(undistorted[side])):
                total_error += abs(undistorted[this_side][i][0][0] *
                                   lines[other_side][i][0][0] +
                                   undistorted[this_side][i][0][1] *
                                   lines[other_side][i][0][1] +
                                   lines[other_side][i][0][2])
            other_side, this_side = sides
        total_points = self.image_count * len(self.object_points)
        return total_error / total_points



# Copyright (C) 2014 Daniel Lee <lee.daniel.1986@gmail.com>
#
# This file is part of StereoVision.
#
# StereoVision is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# StereoVision is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with StereoVision.  If not, see <http://www.gnu.org/licenses/>.

"""
Utilities for 3d reconstruction using stereo cameras.

Modules:

    * ``stereo_cameras`` - Camera interfaces
    * ``calibration`` - Tools for calibrating stereo cameras
    * ``blockmatchers`` - Blockmatching algorithm matchers
    * ``point_cloud`` - Point clouds
    * ``ui_utils`` - Utilities for user interaction
    * ``exceptions`` - Various exceptions

Import structure:

.. image:: packages_StereoVision.svg
    :width: 100%

Camera interfaces
*****************

.. automodule:: stereovision.stereo_cameras

Camera calibration
******************

.. automodule:: stereovision.calibration

Block matchers
**************

.. automodule:: stereovision.blockmatchers

Point clouds
************

.. automodule:: stereovision.point_cloud

User interface utilities
************************

.. automodule:: stereovision.ui_utils

Exceptions
**********

.. automodule:: stereovision.exceptions
"""



# Copyright (C) 2014 Daniel Lee <lee.daniel.1986@gmail.com>
#
# This file is part of StereoVision.
#
# StereoVision is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# StereoVision is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with StereoVision.  If not, see <http://www.gnu.org/licenses/>.

'''
Wrapper classes for block matching algorithms.

Classes:

    * ``BlockMatcher`` - Abstract class that implements interface for subclasses

        * ``StereoBM`` - StereoBM block matching algorithm
        * ``StereoSGBM`` - StereoSGBM block matching algorithm

.. image:: classes_blockmatchers.svg
'''

import cv2
import simplejson

import numpy as np
from stereovision.exceptions import (InvalidSearchRangeError,
                                    InvalidWindowSizeError,
                                    InvalidBMPresetError,
                                    InvalidNumDisparitiesError,
                                    InvalidSADWindowSizeError,
                                    InvalidUniquenessRatioError,
                                    InvalidSpeckleWindowSizeError,
                                    InvalidSpeckleRangeError,
                                    InvalidFirstDisparityChangePenaltyError,
                                    InvalidSecondDisparityChangePenaltyError)


class BlockMatcher(object):

    """
    Block matching algorithms.

    This abstract class exposes the interface for subclasses that wrap OpenCV's
    block matching algorithms. Doing so makes it possible to use them in the
    strategy pattern. In this library, that happens in ``CalibratedPair``, which
    uses a unified interface to interact with any kind of block matcher, and
    with ``BMTuners``, which can discover the ``BlockMatcher's`` parameters and
    allow the user to adjust them online.

    Each ``BlockMatcher`` protects its block matcher's parameters by using
    getters and setters. It exposes its settable parameter and their maximum
    values, if they exist, in the dictionary ``parameter_maxima``.

    ``load_settings``, ``save_settings`` and ``get_3d`` are implemented on
    ``BlockMatcher`` itself, as these are independent of the block matching
    algorithm. Subclasses are expected to implement ``_replace_bm`` and
    ``get_disparity``, as well as the getters and setters. They are also
    expected to call ``BlockMatcher``'s ``__init__`` after setting their own
    private variables.
    """

    #: Dictionary of parameter names associated with their maximum values
    parameter_maxima = {}

    def __init__(self, settings=None):
        """Set block matcher parameters and load from file if necessary."""
        #: Block matcher object used for computing point clouds
        self._block_matcher = None
        self._replace_bm()
        if settings:
            self.load_settings(settings)

    def load_settings(self, settings):
        """Load settings from file"""
        with open(settings) as settings_file:
            settings_dict = simplejson.load(settings_file)
        for key, value in settings_dict.items():
            self.__setattr__(key, value)

    def save_settings(self, settings_file):
        """Save block matcher settings to a file object"""
        settings = {}
        for parameter in self.parameter_maxima:
            settings[parameter] = self.__getattribute__(parameter)
        with open(settings_file, "w") as settings_file:
            simplejson.dump(settings, settings_file)

    @classmethod
    def get_3d(cls, disparity, disparity_to_depth_map):
        """Compute point cloud."""
        return cv2.reprojectImageTo3D(disparity, disparity_to_depth_map)

    def _replace_bm(self):
        """Replace block matcher with new parameters"""
        raise NotImplementedError

    def get_disparity(self, image_pair):
        """Compute disparity map from image pair."""
        raise NotImplementedError


class StereoBM(BlockMatcher):

    """A stereo block matching ``BlockMatcher``."""

    parameter_maxima = {"search_range": None,
                       "window_size": 255,
                       "stereo_bm_preset": cv2.STEREO_BM_NARROW_PRESET}

    @property
    def search_range(self):
        """Return private ``_search_range`` value."""
        return self._search_range

    @search_range.setter
    def search_range(self, value):
        """Set private ``_search_range`` and reset ``_block_matcher``."""
        if value == 0 or not value % 16:
            self._search_range = value
        else:
            raise InvalidSearchRangeError("Search range must be a multiple of "
                                          "16.")
        self._replace_bm()

    @property
    def window_size(self):
        """Return private ``_window_size`` value."""
        return self._window_size

    @window_size.setter
    def window_size(self, value):
        """Set private ``_window_size`` and reset ``_block_matcher``."""
        if (value > 4 and
            value < self.parameter_maxima["window_size"] and
            value % 2):
            self._window_size = value
        else:
            raise InvalidWindowSizeError("Window size must be an odd number "
                                      "between 0 and {}.".format(
                                      self.parameter_maxima["window_size"] + 1))
        self._replace_bm()

    @property
    def stereo_bm_preset(self):
        """Return private ``_bm_preset`` value."""
        return self._bm_preset

    @stereo_bm_preset.setter
    def stereo_bm_preset(self, value):
        """Set private ``_stereo_bm_preset`` and reset ``_block_matcher``."""
        if value in (cv2.STEREO_BM_BASIC_PRESET,
                     cv2.STEREO_BM_FISH_EYE_PRESET,
                     cv2.STEREO_BM_NARROW_PRESET):
            self._bm_preset = value
        else:
            raise InvalidBMPresetError("Stereo BM preset must be defined as "
                                       "cv2.STEREO_BM_*_PRESET.")
        self._replace_bm()

    def _replace_bm(self):
        """Replace ``_block_matcher`` with current values."""
        self._block_matcher = cv2.StereoBM(preset=self._bm_preset,
                                          ndisparities=self._search_range,
                                          SADWindowSize=self._window_size)

    def __init__(self, stereo_bm_preset=cv2.STEREO_BM_BASIC_PRESET,
                 search_range=80,
                 window_size=21,
                 settings=None):
        self._bm_preset = cv2.STEREO_BM_BASIC_PRESET
        self._search_range = 0
        self._window_size = 5
        #: OpenCV camera type for ``_block_matcher``
        self.stereo_bm_preset = stereo_bm_preset
        #: Number of disparities for ``_block_matcher``
        self.search_range = search_range
        #: Search window size for ``_block_matcher``
        self.window_size = window_size
        super(StereoBM, self).__init__(settings)

    def get_disparity(self, pair):
        """
        Compute disparity from image pair (left, right).

        First, convert images to grayscale if needed. Then pass to the
        ``_block_matcher`` for stereo matching.
        """
        gray = []
        if pair[0].ndim == 3:
            for side in pair:
                gray.append(cv2.cvtColor(side, cv2.COLOR_BGR2GRAY))
        else:
            gray = pair
        return self._block_matcher.compute(gray[0], gray[1],
                                          disptype=cv2.CV_32F)


class StereoSGBM(BlockMatcher):

    """A semi-global block matcher."""

    parameter_maxima = {"minDisparity": None,
                       "numDisparities": None,
                       "SADWindowSize": 11,
                       "P1": None,
                       "P2": None,
                       "disp12MaxDiff": None,
                       "uniquenessRatio": 15,
                       "speckleWindowSize": 200,
                       "speckleRange": 2,
                       "fullDP": 1}

    @property
    def minDisparity(self):
        """Return private ``_min_disparity`` value."""
        return self._min_disparity

    @minDisparity.setter
    def minDisparity(self, value):
        """Set private ``_min_disparity`` and reset ``_block_matcher``."""
        self._min_disparity = value
        self._replace_bm()

    @property
    def numDisparities(self):
        """Return private ``_num_disp`` value."""
        return self._num_disp

    @numDisparities.setter
    def numDisparities(self, value):
        """Set private ``_num_disp`` and reset ``_block_matcher``."""
        if value > 0 and value % 16 == 0:
            self._num_disp = value
        else:
            raise InvalidNumDisparitiesError("numDisparities must be a "
                                             "positive integer evenly "
                                             "divisible by 16.")
        self._replace_bm()

    @property
    def SADWindowSize(self):
        """Return private ``_sad_window_size`` value."""
        return self._sad_window_size

    @SADWindowSize.setter
    def SADWindowSize(self, value):
        """Set private ``_sad_window_size`` and reset ``_block_matcher``."""
        if value >= 1 and value <= 11 and value % 2:
            self._sad_window_size = value
        else:
            raise InvalidSADWindowSizeError("SADWindowSize must be odd and "
                                            "between 1 and 11.")
        self._replace_bm()

    @property
    def uniquenessRatio(self):
        """Return private ``_uniqueness`` value."""
        return self._uniqueness

    @uniquenessRatio.setter
    def uniquenessRatio(self, value):
        """Set private ``_uniqueness`` and reset ``_block_matcher``."""
        if value >= 5 and value <= 15:
            self._uniqueness = value
        else:
            raise InvalidUniquenessRatioError("Uniqueness ratio must be "
                                              "between 5 and 15.")
        self._replace_bm()

    @property
    def speckleWindowSize(self):
        """Return private ``_speckle_window_size`` value."""
        return self._speckle_window_size

    @speckleWindowSize.setter
    def speckleWindowSize(self, value):
        """Set private ``_speckle_window_size`` and reset ``_block_matcher``."""
        if value >= 0 and value <= 200:
            self._speckle_window_size = value
        else:
            raise InvalidSpeckleWindowSizeError("Speckle window size must be 0 "
                                                "for disabled checks or "
                                                "between 50 and 200.")
        self._replace_bm()

    @property
    def speckleRange(self):
        """Return private ``_speckle_range`` value."""
        return self._speckle_range

    @speckleRange.setter
    def speckleRange(self, value):
        """Set private ``_speckle_range`` and reset ``_block_matcher``."""
        if value >= 0:
            self._speckle_range = value
        else:
            raise InvalidSpeckleRangeError("Speckle range cannot be negative.")
        self._replace_bm()

    @property
    def disp12MaxDiff(self):
        """Return private ``_max_disparity`` value."""
        return self._max_disparity

    @disp12MaxDiff.setter
    def disp12MaxDiff(self, value):
        """Set private ``_max_disparity`` and reset ``_block_matcher``."""
        self._max_disparity = value
        self._replace_bm()

    @property
    def P1(self):
        """Return private ``_P1`` value."""
        return self._P1

    @P1.setter
    def P1(self, value):
        """Set private ``_P1`` and reset ``_block_matcher``."""
        if value < self.P2:
            self._P1 = value
        else:
            raise InvalidFirstDisparityChangePenaltyError("P1 must be less "
                                                          "than P2.")
        self._replace_bm()

    @property
    def P2(self):
        """Return private ``_P2`` value."""
        return self._P2

    @P2.setter
    def P2(self, value):
        """Set private ``_P2`` and reset ``_block_matcher``."""
        if value > self.P1:
            self._P2 = value
        else:
            raise InvalidSecondDisparityChangePenaltyError("P2 must be greater "
                                                          "than P1.")
        self._replace_bm()

    @property
    def fullDP(self):
        """Return private ``_full_dp`` value."""
        return self._full_dp

    @fullDP.setter
    def fullDP(self, value):
        """Set private ``_full_dp`` and reset ``_block_matcher``."""
        self._full_dp = bool(value)
        self._replace_bm()

    def _replace_bm(self):
        """Replace ``_block_matcher`` with current values."""
        self._block_matcher = cv2.StereoSGBM(minDisparity=self._min_disparity,
                        numDisparities=self._num_disp,
                        SADWindowSize=self._sad_window_size,
                        uniquenessRatio=self._uniqueness,
                        speckleWindowSize=self._speckle_window_size,
                        speckleRange=self._speckle_range,
                        disp12MaxDiff=self._max_disparity,
                        P1=self._P1,
                        P2=self._P2,
                        fullDP=self._full_dp)

    def __init__(self, min_disparity=16, num_disp=96, sad_window_size=3,
                 uniqueness=10, speckle_window_size=100, speckle_range=32,
                 p1=216, p2=864, max_disparity=1, full_dp=False,
                 settings=None):
        """Instantiate private variables and call superclass initializer."""
        #: Minimum number of disparities. Normally 0, can be adjusted as
        #: needed
        self._min_disparity = min_disparity
        #: Number of disparities
        self._num_disp = num_disp
        #: Matched block size
        self._sad_window_size = sad_window_size
        #: Uniqueness ratio for found matches
        self._uniqueness = uniqueness
        #: Maximum size of smooth disparity regions to invalid by noise
        self._speckle_window_size = speckle_window_size
        #: Maximum disparity range within connected component
        self._speckle_range = speckle_range
        #: Penalty on disparity change by +-1 between neighbor pixels
        self._P1 = p1
        #: Penalty on disparity change by multiple neighbor pixels
        self._P2 = p2
        #: Maximum left-right disparity. 0 to disable check
        self._max_disparity = max_disparity
        #: Boolean to use full-scale two-pass dynamic algorithm
        self._full_dp = full_dp
        #: StereoSGBM whose state is controlled
        self._block_matcher = cv2.StereoSGBM()
        super(StereoSGBM, self).__init__(settings)

    def get_disparity(self, pair):
        """Compute disparity from image pair (left, right)."""
        return self._block_matcher.compute(pair[0],
                                          pair[1]).astype(np.float32) / 16.0



# Copyright (C) 2014 Daniel Lee <lee.daniel.1986@gmail.com>
#
# This file is part of StereoVision.
#
# StereoVision is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# StereoVision is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with StereoVision.  If not, see <http://www.gnu.org/licenses/>.

"""
Various exceptions for working with stereovision.

Classes:

    * ``ChessboardNotFoundError``

    * ``BadBlockMatcherArgumentError``

        * ``StereoBMError``

            * ``InvalidBMPresetError``
            * ``InvalidSearchRangeError``
            * ``InvalidWindowSizeError``

        * ``StereoSGBMError``

            * ``InvalidNumDisparitiesError``
            * ``InvalidSADWindowSizeError``
            * ``InvalidFirstDisparityChangePenaltyError``
            * ``InvalidSecondDisparityChangePenaltyError``
            * ``InvalidUniquenessRatioError``
            * ``InvalidSpeckleWindowSizeError``
            * ``InvalidSpeckleRangeError``

.. image:: classes_exceptions.svg
    :width: 100%
"""

class ChessboardNotFoundError(Exception):
    """No chessboard could be found in searched image."""


class BadBlockMatcherArgumentError(Exception):
    """Bad argument supplied for a ``BlockMatcher``."""

class StereoBMError(BadBlockMatcherArgumentError):
    """Bad argument supplied for a ``StereoBM``."""

class StereoSGBMError(BadBlockMatcherArgumentError):
    """Bad argument supplied for a ``StereoSGBM``."""

class InvalidBMPresetError(StereoBMError):
    """Invalid BM preset."""

class InvalidSearchRangeError(StereoBMError):
    """Invalid search range."""

class InvalidWindowSizeError(StereoBMError):
    """Invalid search range."""

class InvalidNumDisparitiesError(StereoSGBMError):
    """Invalid number of disparities."""

class InvalidSADWindowSizeError(StereoSGBMError):
    """Invalid search window size."""

class InvalidFirstDisparityChangePenaltyError(StereoSGBMError):
    """Invalid first disparity change penalty."""

class InvalidSecondDisparityChangePenaltyError(StereoSGBMError):
    """Invalid second disparity change penalty."""

class InvalidUniquenessRatioError(StereoSGBMError):
    """Invalid uniqueness ratio."""

class InvalidSpeckleWindowSizeError(StereoSGBMError):
    """Invalid speckle window size."""

class InvalidSpeckleRangeError(StereoSGBMError):
    """Invalid speckle range."""



# Copyright (C) 2014 Daniel Lee <lee.daniel.1986@gmail.com>
#
# This file is part of StereoVision.
#
# StereoVision is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# StereoVision is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with StereoVision.  If not, see <http://www.gnu.org/licenses/>.

'''
Point cloud class generated from stereo image pairs.

Classes:

    * ``PointCloud`` - Point cloud with RGB colors

.. image:: classes_point_cloud.svg
'''

import numpy as np


class PointCloud(object):

    """3D point cloud generated from a stereo image pair."""

    #: Header for exporting point cloud to PLY
    ply_header = (
'''ply
format ascii 1.0
element vertex {vertex_count}
property float x
property float y
property float z
property uchar red
property uchar green
property uchar blue
end_header
''')

    def __init__(self, coordinates, colors):
        """
        Initialize point cloud with given coordinates and associated colors.

        ``coordinates`` and ``colors`` should be numpy arrays of the same
        length, in which ``coordinates`` is made of three-dimensional point
        positions (X, Y, Z) and ``colors`` is made of three-dimensional spectral
        data, e.g. (R, G, B).
        """
        self.coordinates = coordinates.reshape(-1, 3)
        self.colors = colors.reshape(-1, 3)

    def write_ply(self, output_file):
        """Export ``PointCloud`` to PLY file for viewing in MeshLab."""
        points = np.hstack([self.coordinates, self.colors])
        with open(output_file, 'w') as outfile:
            outfile.write(self.ply_header.format(
                                            vertex_count=len(self.coordinates)))
            np.savetxt(outfile, points, '%f %f %f %d %d %d')

    def filter_infinity(self):
        """Filter infinite distances from ``PointCloud.``"""
        mask = self.coordinates[:, 2] > self.coordinates[:, 2].min()
        coords = self.coordinates[mask]
        colors = self.colors[mask]
        return PointCloud(coords, colors)



# Copyright (C) 2014 Daniel Lee <lee.daniel.1986@gmail.com>
#
# This file is part of StereoVision.
#
# StereoVision is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# StereoVision is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with StereoVision.  If not, see <http://www.gnu.org/licenses/>.

"""
Classes for interacting with stereo cameras.

Classes:

    * ``StereoPair`` - Base class for interacting with stereo cameras

        * ``ChessboardFinder`` - Class for finding chessboards with both cameras
        * ``CalibratedPair`` - Calibrated stereo camera pair that rectifies its
          images

.. image:: classes_stereo_cameras.svg
"""

import cv2

from stereovision.point_cloud import PointCloud


class StereoPair(object):

    """
    A stereo pair of cameras.

    This class allows both cameras in a stereo pair to be accessed
    simultaneously. It also allows the user to show single frames or videos
    captured online with the cameras. It should be instantiated with a context
    manager to ensure that the cameras are freed properly after use.
    """

    #: Window names for showing captured frame from each camera
    windows = ["{} camera".format(side) for side in ("Left", "Right")]

    def __init__(self, devices):
        """
        Initialize cameras.

        ``devices`` is an iterable containing the device numbers.
        """
        #: Video captures associated with the ``StereoPair``
        self.captures = [cv2.VideoCapture(device) for device in devices]

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        for capture in self.captures:
            capture.release()
        for window in self.windows:
            cv2.destroyWindow(window)

    def get_frames(self):
        """Get current frames from cameras."""
        return [capture.read()[1] for capture in self.captures]

    def show_frames(self, wait=0):
        """
        Show current frames from cameras.

        ``wait`` is the wait interval in milliseconds before the window closes.
        """
        for window, frame in zip(self.windows, self.get_frames()):
            cv2.imshow(window, frame)
        cv2.waitKey(wait)

    def show_videos(self):
        """Show video from cameras."""
        while True:
            self.show_frames(1)
            if cv2.waitKey(1) & 0xFF == ord('q'):
                break


class ChessboardFinder(StereoPair):

    """A ``StereoPair`` that can find chessboards in its images."""

    def get_chessboard(self, columns, rows, show=False):
        """
        Take a picture with a chessboard visible in both captures.

        ``columns`` and ``rows`` should be the number of inside corners in the
        chessboard's columns and rows. ``show`` determines whether the frames
        are shown while the cameras search for a chessboard.
        """
        found_chessboard = [False, False]
        while not all(found_chessboard):
            frames = self.get_frames()
            if show:
                self.show_frames(1)
            for i, frame in enumerate(frames):
                (found_chessboard[i],
                 corners) = cv2.findChessboardCorners(frame, (columns, rows),
                                                  flags=cv2.CALIB_CB_FAST_CHECK)
        return frames


class CalibratedPair(StereoPair):

    """
    A ``StereoPair`` that works with rectified images and produces point clouds.
    """

    def __init__(self, devices, calibration, block_matcher):
        """
        Initialize cameras.

        ``devices`` is an iterable of the device numbers. If you want to use the
        ``CalibratedPair`` in offline mode, it should be None.
        ``calibration`` is a ``StereoCalibration`` object.
        ``block_matcher`` is a ``BlockMatcher`` object.
        """
        if devices:
            super(CalibratedPair, self).__init__(devices)
        #: ``StereoCalibration`` object holding the camera pair's calibration
        self.calibration = calibration
        #: ``BlockMatcher`` object for computing disparity and point cloud
        self.block_matcher = block_matcher

    def get_frames(self):
        """Rectify and return current frames from cameras."""
        frames = super(CalibratedPair, self).get_frames()
        return self.calibration.rectify(frames)

    def get_point_cloud(self, pair):
        """Get 3D point cloud from image pair."""
        disparity = self.block_matcher.get_disparity(pair)
        points = self.block_matcher.get_3d(disparity,
                                           self.calibration.disp_to_depth_mat)
        colors = cv2.cvtColor(pair[0], cv2.COLOR_BGR2RGB)
        return PointCloud(points, colors)



# Copyright (C) 2014 Daniel Lee <lee.daniel.1986@gmail.com>
#
# This file is part of StereoVision.
#
# StereoVision is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# StereoVision is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with StereoVision.  If not, see <http://www.gnu.org/licenses/>.

from setuptools import setup

setup(name="StereoVision",
      version="1.0.1",
      description=("Library and utilities for 3d reconstruction from stereo "
                   "cameras."),
      long_description=open("README.rst").read(),
      author="Daniel Lee",
      author_email="lee.daniel.1986@gmail.com",
      packages=["stereovision"],
      scripts=["bin/calibrate_cameras",
               "bin/capture_chessboards",
               "bin/images_to_pointcloud",
               "bin/show_webcams",
               "bin/tune_blockmatcher"],
      url="http://erget.github.com/StereoVision",
      download_url="http://pypi.python.org/pypi/StereoVision",
      license="GNU GPL",
      requires=["cv2",
                "simplejson",
                "numpy",
                "progressbar"],
      provides=["stereovision"],
      classifiers=["Development Status :: 5 - Production/Stable",
                   "Natural Language :: English",
                   "Operating System :: OS Independent",
                   "Programming Language :: Python :: 2",
                   "Intended Audience :: Developers",
                   "Intended Audience :: Education",
                   "Intended Audience :: Science/Research",
                   "License :: Freely Distributable",
                   "License :: OSI Approved :: GNU General Public License v3 "
                                              "or later (GPLv3+)",
                   "Natural Language :: English",
                   "Operating System :: OS Independent",
                   "Programming Language :: Python :: 2.7",
                   "Topic :: Multimedia :: Graphics :: Capture"])



# Copyright (C) 2014 Daniel Lee <lee.daniel.1986@gmail.com>
#
# This file is part of StereoVision.
#
# StereoVision is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# StereoVision is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with StereoVision.  If not, see <http://www.gnu.org/licenses/>.

"""
Utilities for easing user interaction with the ``stereovision`` package.

Variables:

    * ``CHESSBOARD_ARGUMENTS`` - ``argparse.ArgumentParser`` for working with
      chessboards
    * ``STEREO_BM_FLAG`` - ``argparse.ArgumentParser`` for using StereoBM

Functions:

    * ``find_files`` - Discover stereo images in directory
    * ``calibrate_folder`` - Calibrate chessboard images discoverd in a folder

Classes:

    * ``BMTuner`` - Tune block matching algorithm to camera pair

.. image:: classes_ui_utils.svg
"""

from argparse import ArgumentParser
from functools import partial
import os

import cv2
from progressbar import ProgressBar, Percentage, Bar
from stereovision.calibration import StereoCalibrator
from stereovision.exceptions import BadBlockMatcherArgumentError

#: Command line arguments for collecting information about chessboards
CHESSBOARD_ARGUMENTS = ArgumentParser(add_help=False)
CHESSBOARD_ARGUMENTS.add_argument("--rows", type=int,
                                  help="Number of inside corners in the "
                                  "chessboard's rows.", default=9)
CHESSBOARD_ARGUMENTS.add_argument("--columns", type=int,
                                  help="Number of inside corners in the "
                                  "chessboard's columns.", default=6)
CHESSBOARD_ARGUMENTS.add_argument("--square-size", help="Size of chessboard "
                                  "squares in cm.", type=float, default=1.8)


#: Command line arguments for using StereoBM rather than StereoSGBM
STEREO_BM_FLAG = ArgumentParser(add_help=False)
STEREO_BM_FLAG.add_argument("--use_stereobm", help="Use StereoBM rather than "
                              "StereoSGBM block matcher.", action="store_true")


def find_files(folder):
    """Discover stereo photos and return them as a pairwise sorted list."""
    files = [i for i in os.listdir(folder) if i.startswith("left")]
    files.sort()
    for i in range(len(files)):
        insert_string = "right{}".format(files[i * 2][4:])
        files.insert(i * 2 + 1, insert_string)
    files = [os.path.join(folder, filename) for filename in files]
    return files


def calibrate_folder(args):
    """
    Calibrate camera based on chessboard images, write results to output folder.

    All images are read from disk. Chessboard points are found and used to
    calibrate the stereo pair. Finally, the calibration is written to the folder
    specified in ``args``.

    ``args`` needs to contain the following fields:
        input_files: List of paths to input files
        rows: Number of rows in chessboard
        columns: Number of columns in chessboard
        square_size: Size of chessboard squares in cm
        output_folder: Folder to write calibration to
    """
    height, width = cv2.imread(args.input_files[0]).shape[:2]
    calibrator = StereoCalibrator(args.rows, args.columns, args.square_size,
                                  (width, height))
    progress = ProgressBar(maxval=len(args.input_files),
                          widgets=[Bar("=", "[", "]"),
                          " ", Percentage()])
    print("Reading input files...")
    progress.start()
    while args.input_files:
        left, right = args.input_files[:2]
        img_left, im_right = cv2.imread(left), cv2.imread(right)
        calibrator.add_corners((img_left, im_right),
                               show_results=args.show_chessboards)
        args.input_files = args.input_files[2:]
        progress.update(progress.maxval - len(args.input_files))

    progress.finish()
    print("Calibrating cameras. This can take a while.")
    calibration = calibrator.calibrate_cameras()
    avg_error = calibrator.check_calibration(calibration)
    print("The average error between chessboard points and their epipolar "
          "lines is \n"
          "{} pixels. This should be as small as possible.".format(avg_error))
    calibration.export(args.output_folder)


class BMTuner(object):

    """
    A class for tuning Stereo BM settings.

    Display a normalized disparity picture from two pictures captured with a
    ``CalibratedPair`` and allow the user to manually tune the settings for the
    ``BlockMatcher``.

    The settable parameters are intelligently read from the ``BlockMatcher``,
    relying on the ``BlockMatcher`` exposing them as ``parameter_maxima``.
    """

    #: Window to show results in
    window_name = "BM Tuner"

    def _set_value(self, parameter, new_value):
        """Try setting new parameter on ``block_matcher`` and update map."""
        try:
            self.block_matcher.__setattr__(parameter, new_value)
        except BadBlockMatcherArgumentError:
            return
        self.update_disparity_map()

    def _initialize_trackbars(self):
        """
        Initialize trackbars by discovering ``block_matcher``'s parameters.
        """
        for parameter in self.block_matcher.parameter_maxima.keys():
            maximum = self.block_matcher.parameter_maxima[parameter]
            if not maximum:
                maximum = self.shortest_dimension
            cv2.createTrackbar(parameter, self.window_name,
                               self.block_matcher.__getattribute__(parameter),
                               maximum,
                               partial(self._set_value, parameter))

    def _save_bm_state(self):
        """Save current state of ``block_matcher``."""
        for parameter in self.block_matcher.parameter_maxima.keys():
            self.bm_settings[parameter].append(
                               self.block_matcher.__getattribute__(parameter))

    def __init__(self, block_matcher, calibration, image_pair):
        """
        Initialize tuner window and tune given pair.

        ``block_matcher`` is a ``BlockMatcher``, ``calibration`` is a
        ``StereoCalibration`` and ``image_pair`` is a rectified image pair.
        """
        #: Stereo calibration to find Stereo BM settings for
        self.calibration = calibration
        #: (left, right) image pair to find disparity between
        self.pair = image_pair
        #: Block matcher to be tuned
        self.block_matcher = block_matcher
        #: Shortest dimension of image
        self.shortest_dimension = min(self.pair[0].shape[:2])
        #: Settings chosen for ``BlockMatcher``
        self.bm_settings = {}
        for parameter in self.block_matcher.parameter_maxima.keys():
            self.bm_settings[parameter] = []
        cv2.namedWindow(self.window_name)
        self._initialize_trackbars()
        self.tune_pair(image_pair)

    def update_disparity_map(self):
        """
        Update disparity map in GUI.

        The disparity image is normalized to the range 0-255 and then divided by
        255, because OpenCV multiplies it by 255 when displaying. This is
        because the pixels are stored as floating points.
        """
        disparity = self.block_matcher.get_disparity(self.pair)
        norm_coeff = 255 / disparity.max()
        cv2.imshow(self.window_name, disparity * norm_coeff / 255)
        cv2.waitKey()

    def tune_pair(self, pair):
        """Tune a pair of images."""
        self._save_bm_state()
        self.pair = pair
        self.update_disparity_map()

    def report_settings(self, parameter):
        """
        Report chosen settings for ``parameter`` in ``block_matcher``.

        ``bm_settings`` is updated to include the latest state before work is
        begun. This state is removed at the end so that the method has no side
        effects. All settings are reported except for the first one on record,
        which is ``block_matcher``'s default setting.
        """
        self._save_bm_state()
        report = []
        settings_list = self.bm_settings[parameter][1:]
        unique_values = list(set(settings_list))
        value_frequency = {}
        for value in unique_values:
            value_frequency[settings_list.count(value)] = value
        frequencies = value_frequency.keys()
        frequencies.sort(reverse=True)
        header = "{} value | Selection frequency".format(parameter)
        left_column_width = len(header[:-21])
        right_column_width = 21
        report.append(header)
        report.append("{}|{}".format("-" * left_column_width,
                                    "-" * right_column_width))
        for frequency in frequencies:
            left_column = str(value_frequency[frequency]).center(
                                                             left_column_width)
            right_column = str(frequency).center(right_column_width)
            report.append("{}|{}".format(left_column, right_column))
        # Remove newest settings
        for param in self.block_matcher.parameter_maxima.keys():
            self.bm_settings[param].pop(-1)
        return "\n".join(report)
